/*
  CommBuffer.cxx
  --------------
  WritePointer and ReadPointer points to the address where it will act on, 
  on the next request.
*/
#include <iostream>
#include <algorithm>
#include <cstring>
#include <cstdio>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "CommBase/CommBuffer.hxx"

CommBuffer::CommBuffer() {
  mBufferSize = 0;
  mData = 0;
  mWritePointer = 0;
  mReadPointer = 0;
}

CommBuffer::~CommBuffer() {
  free_data();
}

int CommBuffer::allocate_data(unsigned int nbytes) {
  free_data();
  mData = new char[nbytes];
  mBufferSize = nbytes;
  return 0;
}

int CommBuffer::free_data() {
  if (mData) {
    delete [] mData;
    mData = 0;
  }
  mBufferSize = 0;
  return 0;
}

bool CommBuffer::empty() const {
  bool s = false;
  if (mWritePointer == mReadPointer) {
    s = true;
  } else {
    s = false;
  }
  return s;
}

unsigned int CommBuffer::sizeForRead() const {
  unsigned int n=0;
  if (mWritePointer > mReadPointer) {
    n = mWritePointer - mReadPointer;
  } else if (mWritePointer < mReadPointer) {
    n = mBufferSize - mReadPointer + mWritePointer;
  } else {
    n = 0;
  }
  return n;
}

unsigned int CommBuffer::sizeForReadContinuous() const {
  unsigned int n=0;
  if (mReadPointer < mWritePointer) {
    n = mWritePointer - mReadPointer;
  } else  if (mWritePointer < mReadPointer) {
    n = mBufferSize - mReadPointer;
  } else {
    n = 0;
  }
  return n;
}

unsigned int CommBuffer::sizeForWrite() const {
  unsigned int n=0;
  if (mWritePointer > mReadPointer) {
    n = mBufferSize - mWritePointer + mReadPointer;
  } else if (mWritePointer < mReadPointer) {
    n = mReadPointer - mWritePointer;
  } else {
    n = mBufferSize;
  }
  return n;
}

unsigned int CommBuffer::sizeForWriteContinuous() const {
  unsigned int n=0;
  if (mWritePointer >= mReadPointer) {
    n = mBufferSize - mWritePointer;
  } else { // mWritePointer < mReadPointer
    n = mReadPointer - mWritePointer;
  }
  return n;
}

int CommBuffer::read(void* buf, unsigned int n) {
  unsigned int n_left;
  int n_done;
  unsigned int n1, n2;
  int ntry=0;
  int max_try=1000;

  // std::cout << "Buffer read: available=" << sizeForRead() 
  // 	    << ", n(req)=" << n << std::endl;
  // print();
  // Cache incoming data in the buffer
  while (sizeForRead() < n) {
    if (ntry > 2) {
      usleep(1);
    }
    n1 = sizeForWriteContinuous();
    n2 = recv_data(mData+mWritePointer, n1);
    mWritePointer = (mWritePointer + n2) % mBufferSize;

    if (ntry >= max_try) break;
    ntry ++;
  }
  // std::cout << "read into cache: " << sizeForRead() << std::endl;
  // print();

  // Copy data from the internal buffer to the argument (buf)
  n_left = n;
  n_done = 0;
  if (n <= sizeForRead()) {
    for (int i=0; i<2; ++i) {
      n1 = std::min(sizeForReadContinuous(), n_left);
      std::memcpy(buf, mData+mReadPointer, n1);
      mReadPointer = (mReadPointer + n1) % mBufferSize;
      n_done += n1;
      n_left -= n1;
    }
  }
  return n_done;
}

int CommBuffer::write(const void* buf, unsigned int n) {
  return send_data(buf, n);
}

const std::string CommBuffer::socketType() const {
  return "TCP";
}

void CommBuffer::setIPAddressPort(const std::string& addr, int port) {
  mIPAddress = addr;
  mPort = port;
}

void CommBuffer::updateDestinationFromSenderAddress() {
}

void CommBuffer::print() const {
  std::printf("CommBuffer [type=%s, socket=%d]: \n", 
	      socketType().c_str(), mSocket);
  std::printf("  Buffer size   : %8d\n", mBufferSize);
  std::printf("  Read pointer  : %8d\n", mReadPointer);
  std::printf("  Write pointer : %8d\n", mWritePointer);
  std::printf("  Size for read : %8d, continuous=%8d\n", 
	      sizeForRead(), sizeForReadContinuous());
  std::printf("  Size for write: %8d, continuous=%8d\n", 
	      sizeForWrite(), sizeForWriteContinuous());
}

int CommBuffer::recv_data(void* buf, unsigned int n) {
  int flags = MSG_DONTWAIT;
  flags = 0;
  return ::recv(mSocket, buf, n, flags);
  //return ::read(mSocket, buf, n);
}

int CommBuffer::send_data(const void* buf, unsigned int n) {
  // const char* cbuf = reinterpret_cast<const char*>(buf);
  // std::cout << "Socket in send_data: " << mSocket << std::endl;
  // std::cout << " write " << n << " bytes" << std::endl;
  // for (unsigned int i=0; i<n; ++i) {
  //   std::cout << "  cbuf[" << i << "] " << cbuf[i] << " " 
  // 	      << static_cast<const int>(cbuf[i]) << std::endl;
  // }
  // std::cout << " data=" << cbuf << std::endl;
  // std::cout << " go" << std::endl;
  return ::send(mSocket, buf, n, 0);
}

