/*
  CommunicationMgr.cxx
*/
#include <cstring>
#include <cstdlib>
#include <cstdio>
#include <iostream>
#include <sys/socket.h>
#include <sys/select.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include "CommBase/CommunicationMgr.hxx"
#include "CommBase/GlobalConfig.hxx"
#include "CommBase/CommBuffer.hxx"
#include "CommBase/CommBufferUDP.hxx"

using namespace std;

CommunicationMgr::CommunicationMgr() : 
  mIPAddress(""), mPort(0), mSocket(0), mMessageNumber(0) {
  mSocketType = "TCP";
  mBuffer = 0;
}

CommunicationMgr::CommunicationMgr(const std::string& socket_type, 
				   unsigned int buffer_size) : 
  mIPAddress(""), mPort(0), mSocket(0), mMessageNumber(0){
  mSocketType = socket_type;
  createBuffer(buffer_size, socket_type);
}

CommunicationMgr::~CommunicationMgr() {
  close();
  if (mSubMgrs.size() > 0) {
    std::map<int, CommunicationMgr*>::iterator p;
    for (p=mSubMgrs.begin(); p!=mSubMgrs.end(); ++p) {
      if (p->second != 0) {
	delete p->second;
	p->second = 0;
      }
    }
    mSubMgrs.clear();
  }
  if (mBuffer) {
    delete mBuffer;
    mBuffer = 0;
  }
}

void CommunicationMgr::createBuffer(unsigned int n, 
				    const std::string& socket_type) {
  mSocketType = socket_type;
  std::cout << "Create communication buffer (" << mSocketType << ") size="
	    << n << std::endl;
  if (socket_type == "TCP") {
    mBuffer = new CommBuffer();
  } else if (socket_type == "UDP") {
    mBuffer = new CommBufferUDP();
  } else {
    mBuffer = new CommBuffer();
  }
  mBuffer->allocate_data(n);
}

void CommunicationMgr::setIPAddressPort(const std::string ipaddress, int port) {
  mIPAddress = ipaddress;
  mPort = port;
  mBuffer->setIPAddressPort(ipaddress, port);
}

void CommunicationMgr::setSocket(int fd) {
  if (mSocket > 0) close();
  mSocket = fd;
  if (mBuffer) mBuffer->setSocket(fd);
}

int CommunicationMgr::bind() {
  std::string stype = mSocketType;

  struct addrinfo hints;
  struct addrinfo *result=0;
  struct addrinfo *rp=0;
  memset(&hints, 0, sizeof(struct addrinfo));
  hints.ai_canonname = NULL;
  hints.ai_addr = NULL;
  hints.ai_next = NULL;
  hints.ai_family = AF_INET;
  hints.ai_flags = AI_PASSIVE | AI_NUMERICSERV;
  hints.ai_socktype = SOCK_STREAM;
  if (stype == "TCP") {
    hints.ai_socktype = SOCK_STREAM;
  } else if (stype == "UDP") {
    hints.ai_socktype = SOCK_DGRAM;
  }

  char port_str[10];
  std::sprintf(port_str, "%d", mPort);

  if (getaddrinfo(mIPAddress.c_str(), port_str, &hints, &result) != 0) {
    std::cerr << "Error in getaddrinfo(...)" << std::endl;
    return -1;
  }
  for (rp=result;rp !=NULL; rp=rp->ai_next) {
    mSocket = socket(rp->ai_family, rp->ai_socktype, rp->ai_protocol);
    if (mSocket == -1) {
      continue;
    }
    if (::bind(mSocket, rp->ai_addr, rp->ai_addrlen) != -1) {
      break;
    }
    ::close(mSocket);
  }

  if (rp == NULL) {
    return -3;
  }

  freeaddrinfo(result);

  mBuffer->setSocket(mSocket);

  return 0;
}

int CommunicationMgr::listen() {
  int backlog = 50;
  if (::listen(mSocket, backlog) == -1) {
    std::cerr << "Error while socket listen" << std::endl;
    return -4;
  }
  return 0;
}

CommunicationMgr* CommunicationMgr::accept() {
  struct sockaddr_storage claddr;
  socklen_t addrlen = sizeof(struct sockaddr_storage);
  int cfd = ::accept(mSocket, (struct sockaddr *) &claddr, &addrlen);
  std::cout << "Accepted connection from and created a socket " 
	    << cfd << std::endl;
  CommunicationMgr* c=0;
  if (cfd > 0) {
    c = new CommunicationMgr(mSocketType, 10000);
    c->setSocket(cfd);
    mSubMgrs[cfd] = c;
  }
  //  return cfd;
  return c;
}

int CommunicationMgr::connect() {
  std::string stype=mSocketType;

  struct addrinfo hints;
  struct addrinfo *result=0;
  struct addrinfo *rp=0;
  memset(&hints, 0, sizeof(struct addrinfo));
  hints.ai_canonname = NULL;
  hints.ai_addr = NULL;
  hints.ai_next = NULL;
  hints.ai_family = AF_UNSPEC;
  hints.ai_flags = AI_NUMERICSERV;
  hints.ai_socktype = SOCK_STREAM;
  if (stype == "TCP") {
    hints.ai_socktype = SOCK_STREAM;
  } else if (stype == "UDP") {
    hints.ai_socktype = SOCK_DGRAM;
  }

  char port_str[10];
  std::sprintf(port_str, "%d", mPort);

  if (getaddrinfo(mIPAddress.c_str(), port_str, &hints, &result) != 0) {
    std::cerr << "Error in getaddrinfo(...)" << std::endl;
    return -1;
  }
  for (rp=result;rp !=NULL; rp=rp->ai_next) {
    mSocket = socket(rp->ai_family, rp->ai_socktype, rp->ai_protocol);
    if (mSocket == -1) {
      continue;
    }
    if (::connect(mSocket, rp->ai_addr, rp->ai_addrlen) != -1) {
      break;
    }
    ::close(mSocket);
  }
  if (rp == NULL) {
    std::cerr << "Could not connect socket to any address" << std::endl;
    return -2;
  }
  freeaddrinfo(result);

  mBuffer->setSocket(mSocket);

  return 0;
}

int CommunicationMgr::createUDPSocket() {
  mSocket = socket(AF_INET, SOCK_DGRAM, 0);
  mBuffer->setSocket(mSocket);
  return 0;
}

int CommunicationMgr::close() {
  if (mSocket > 0) ::close(mSocket);
  return 0;
}

int CommunicationMgr::write(CommandSet& cmd) {
  GlobalConfig* gconfig = GlobalConfig::get();

  const char* data = cmd.data();
  int n;
  unsigned int ntot=0;
  usleep(2);

  n = mBuffer->write(data, cmd.packetSize());
  if (n != cmd.packetSize() && gconfig->output(GlobalConfig::kWARNING)) {
    std::cout << "Warning: write(...) " 
  	      << n << " bytes written while trying to write " 
  	      << cmd.packetSize() << " bytes" << endl;
  }
  ntot += n;

  if (gconfig->output(GlobalConfig::kVERBOSE)) {
    std::cout << "CommunicationMgr.write finished, wrote "
	      << ntot << " bytes" << std::endl;
  }
  return ntot;
}

int CommunicationMgr::write(A6Message& msg) {
  GlobalConfig* gconfig = GlobalConfig::get();

  const char* data = msg.data();
  int n;
  unsigned int ntot=0;

  n = mBuffer->write(data, msg.packetSize());
  if (n != msg.packetSize() && gconfig->output(GlobalConfig::kWARNING)) {
    std::cout << "Warning: write(...) " 
  	      << n << " bytes written while trying to write " 
  	      << msg.packetSize() << " bytes" << endl;
  }
  ntot += n;

  if (gconfig->output(GlobalConfig::kVERBOSE)) {
    std::cout << "CommunicationMgr.write finished, wrote "
	      << ntot << " bytes" << std::endl;
  }
  return ntot;
}

int CommunicationMgr::read(CommandSet& cmd) {
  GlobalConfig* gconfig = GlobalConfig::get();

  cmd.clear();
  int status=0;
  unsigned int n=0;
  unsigned int n1=0;
  unsigned int sz=0;
  unsigned int n0=1000;
  char line[1000];
  unsigned int sztot=0;

  if (gconfig->output(GlobalConfig::kVERBOSE)) {
    std::cout << std::endl;
    std::cout << "start reading Message..." << std::endl;
    std::cout << std::endl;
    std::cout << "reading from socket: " << mSocket << std::endl;
    std::cout << "Read message type" << std::endl;
  }

  if ( (n = mBuffer->read(line, 8)) != 8) { 
    if (n == 0) return -1;
    if (gconfig->output(GlobalConfig::kERROR)) {
      std::cerr << "Error while reading CommandSet header returned " << n 
		<< std::endl;
    }
    status = -1;
  } else {
    int p_type=0, p_rw=0, p_cid=0, p_plsz=0;
    unsigned int p_add=0, tmp=0;
    p_type = line[0] & 0xff;
    p_rw = line[1] & 0xff;
    p_cid = line[2] & 0xff;
    p_plsz = line[3] & 0xff;
    for (int i=0; i<4; ++i) {
      tmp = line[4+i] & 0xff;
      p_add |= (tmp<<(24-8*i));
    }
    // std::cout<<"p_type = "<< p_type <<std::endl;
    // std::cout<<"p_rw   = "<< p_rw <<std::endl;
    // std::cout<<"p_cid  = "<< p_cid <<std::endl;
    // std::cout<<"p_plsz = "<< p_plsz <<std::endl;
    // std::cout<<"p_add  = "<< p_add <<std::endl;
    cmd.setType(p_type);
    cmd.setRW(p_rw);
    cmd.setCommandID(p_cid);
    cmd.setPayloadSize(p_plsz);
    cmd.setAddress(p_add);
    sz = p_plsz;
  }
  if (gconfig->output(GlobalConfig::kVERBOSE)) {
    std::cout << "Read message payload (" << sz << ")" << std::endl;
  }
  
  cmd.setPayload("", 0);
  sztot=0;
  while (sztot < sz) {
    n1 = sz - sztot;
    if (n1 > n0) n1 = n0;
    if ( (n = mBuffer->read(line, n1)) != n1) {
      if (n == 0) return -1;
      if (gconfig->output(GlobalConfig::kERROR) ) {
	std::cerr << "Error while reading message payload returned " << n 
		  << ", request was " << n1 << " bytes." << std::endl;
      }
      status = -1;
      break;
    }
    line[n] = '\0';
    sztot += n;
    cmd.appendPayload(line, n);
    if (sztot >= sz) break;
  }

  return status;
}


int CommunicationMgr::read(A6Message& msg) {
  GlobalConfig* gconfig = GlobalConfig::get();

  msg.clear();
  int status=0;
  unsigned int n=0;
  unsigned int n1=0;
  unsigned int sz=0;
  unsigned int n0=1000;
  char line[1000];
  unsigned int sztot=0;
  
  if (gconfig->output(GlobalConfig::kVERBOSE)) {
    std::cout << std::endl;
    std::cout << "start reading Message..." << std::endl;
    std::cout << std::endl;
    std::cout << "reading from socket: " << mSocket << std::endl;
    std::cout << "Read message type" << std::endl;
  }

  if ( (n = mBuffer->read(line, 3)) != 3) { 
    if (n == 0) return -1;
    if (gconfig->output(GlobalConfig::kERROR)) {
      std::cerr << "Error while reading A6Message header returned " << n 
		<< std::endl;
    }
    status = -1;
  } else {
    int p_type=0, p_cid=0, p_plsz=0;
    p_type = line[0] & 0xff;
    p_cid = line[1] & 0xff;
    p_plsz = line[2] & 0xff;
    // std::cout<<"p_type = "<< p_type <<std::endl;
    // std::cout<<"p_cid  = "<< p_cid <<std::endl;
    // std::cout<<"p_plsz = "<< p_plsz <<std::endl;
    msg.setType(p_type);
    msg.setCommandID(p_cid);
    msg.setPayloadSize(p_plsz);
    sz = p_plsz;
  }
  if (gconfig->output(GlobalConfig::kVERBOSE)) {
    std::cout << "Read message payload (" << sz << ")" << std::endl;
  }
  
  msg.setPayload("", 0);
  sztot=0;
  while (sztot < sz) {
    n1 = sz - sztot;
    if (n1 > n0) n1 = n0;
    if ( (n = mBuffer->read(line, n1)) != n1) {
      if (n == 0) return -1;
      if (gconfig->output(GlobalConfig::kERROR) ) {
	std::cerr << "Error while reading message payload returned " << n 
		  << ", request was " << n1 << " bytes." << std::endl;
      }
      status = -1;
      break;
    }
    line[n] = '\0';
    sztot += n;
    msg.appendPayload(line, n);
    if (sztot >= sz) break;
  }

  return status;
}

